import rhinoscriptsyntax as rs
import compas
from compas.datastructures.network import Network
from compas.utilities import geometric_key
from compas.geometry import length_vector_sqrd

__author__    = 'Tomas Mendez Echenagucia'
__copyright__ = 'Copyright 2016, Block Research Group - ETH Zurich'
__license__   = 'MIT license'
__email__     = 'mendez@arch.ethz.ch'

##################################################################################################
# editted for compas 0.16.9 (new definition of network class) by Elizabeta Samec (esamec@grad.hr)
##################################################################################################

__all__ = [
    'make_gkdict',
    'from_lines',
    'update_network_edge_attributes',
    'update_network_vertex_attributes',
    'draw_reactions'
]

def make_gkdict(network, precision='3f'):
    gkdict = {}
    for key in network.node:
        xyz = network.node_attributes(key, ['x', 'y', 'z'])
        gk = geometric_key(xyz=xyz, precision=precision)
        gkdict[gk] = key
    return gkdict

def draw_reactions(network):
    tol = 1e-03
    tol2 = tol ** 2
    vector = []
    for key, attr in network.nodes(True):
        if not attr['is_anchor']:
            continue
        r = rx, ry, rz = network.node_attributes(key, ('rx', 'ry', 'rz'))
        if length_vector_sqrd(r) <= tol2:
            continue
        sp = [x, y, z] = network.node_coordinates(key)
        ep = [x - rx, y - ry, z - rz]
        vector.append(rs.AddLine(sp, ep))
    for key, attr in network.nodes(True):
        if not attr['is_loaded']:
            continue
        r = rx, ry, rz = network.node_attributes(key, ('rx', 'ry', 'rz'))
        if length_vector_sqrd(r) <= tol2:
            continue
        sp = [x, y, z] = network.node_coordinates(key)
        ep = [x - rx, y - ry, z - rz]
        vector.append(rs.AddLine(sp, ep))
    return vector

def from_lines(guids):
    lines = [[rs.CurveStartPoint(g), rs.CurveEndPoint(g)] for g in guids]
    network = Network.from_lines(lines)
    return network

def update_network_edge_attributes(network, lines, name, values):
    gkdict = make_gkdict(network, precision='3f')
    if len(values) == 1:
        values = [values[0]] * len(lines)
    for i, line in enumerate(lines):
        value = values[i]
        u = gkdict[geometric_key(rs.CurveStartPoint(line), precision='3f')]
        v = gkdict[geometric_key(rs.CurveEndPoint(line), precision='3f')]
        key = u,v
        network.edge_attribute(key, name, value)

def update_network_vertex_attributes(network, points, name, values):
    gkdict = make_gkdict(network, precision='3f')
    if len(values) == 1:
        values = [values[0]] * len(points)
    for i, point in enumerate(points):
        value = values[i]
        u = gkdict[geometric_key(point, precision='3f')]
        network.node_attribute(u, name, value)
        
def move_network(network, vector):
    """Move the entire network.

    Parameters:
        network (compas.datastructures.network.Network): A network object.
        vector (list, tupple): The displacement vector.

    """

    for key, attr in network.nodes(True):
        attr['x'] += vector[0]
        attr['y'] += vector[1]
        attr['z'] += vector[2]
    return network


def move_network_vertex(network, points, vectors):
    """Move the entire network.

    Parameters:
        network (compas.datastructures.network.Network): A network object.
        vector (list, tupple): The displacement vector.

    """
    gkdict = make_gkdict(network, precision='3f')
    if len(vectors) == 1:
        vectors = [vectors[0]] * len(points)
    #for i, point in enumerate(points):
    for i, point in points:
        u = gkdict[geometric_key(point, precision='3f')]
        vector = vectors[i]
        network.node[u]['x'] += vector[0]
        network.node[u]['y'] += vector[1]
        network.node[u]['z'] += vector[2]


 
####################################################################################################
## Support for FALCON in Rhino and GH - functions for assigment of constrints (numpy not needed)
#__author__    = ['Kresimir Fresl']
#__email__     = 'fresl@grad.hr'
####################################################################################################

def get_index(edges, keys):
    lis=[]
    for i in keys:
        for a, b in enumerate (edges):
            if b==i:
                lis.append(a)
    return lis

def get_values(lis, lvs):
    return [lvs[i] for i in lis]

def ith_node_coords(ni, vertices):
    return vertices[ni]

def element_node_indices(el):
    return el

def element_node_coords(el, vertices):
    return [ith_node_coords(i, vertices) for i in element_node_indices(el)]

def list_of_element_forces(lengths, q):
    return map(lambda x, y: x * y, lengths, q)

def unstrained_length(l, f, ae):
    return ae * l / (ae + f)

def list_of_unstrained_lengths(ll, ff, ae):
    return map(lambda x, y, z: z * x / (z + y), ll, ff, ae)

def map_value_to_all_edges(edges,v):
    return [v for x in range(len(edges))]

def edge_constraints(ed_indices, ed_values):
    constrs = []
    constrs.extend(zip(ed_indices, ed_values))
    return constrs

def add_edge_constraints(el_indices, el_values, constrs):
    constrs.extend(zip(el_indices, el_values))
    return constrs

def remove_edge_constraints(el_indices, constrs):
    constrs2 = []
    for cnst in constrs:
        ei = cnst[0]
        if ei not in el_indices:
            constrs2.append(cnst)
    return constrs2