from __future__ import division
from __future__ import print_function

import sys
import copy

try:
    from numpy import mean
    from numpy import asarray
    from scipy.sparse import diags
    from scipy.sparse import csc_matrix
    from scipy.linalg import norm as v_norm
    from scipy.sparse.linalg import spsolve
    from scipy.sparse.linalg import norm as sm_norm
    from scipy.sparse.linalg import inv
    from scipy.sparse.linalg import splu
    from scipy.sparse.linalg import cg
    from numpy.linalg import cond
    from numpy import array as array_m
    from numpy import zeros as zero_m
    from numpy import empty as empty_m
    from numpy import asmatrix as as_matrix
    from math import sqrt
    from copy import copy

except ImportError:
    if 'ironpython' not in sys.version.lower():
        raise

#from compas.numerical import normrow

__all__ = [
    'FDM'
    'multistepFDM_d'
    'multistepFDM_it'
    'multistepFDM_wtol_d'
    'multistepFDM_wtol_it'
    'multistepFDM_inexact_grid',
]

__author__    = ['Kresimir Fresl', 'Elizabeta Samec']
__email__     = 'esamec@grad.hr'

no_callbacks = (None, None, None)

##############################################
# Utility functions:
##############################################

def all_of(lst):
    return range(len(lst))

def get_index(edges, keys):
    lis=[]
    for i in keys:
        for a, b in enumerate (edges):
            if b==i:
                lis.append(a)
    return lis

def get_values(lis, lvs):
    return [lvs[i] for i in lis]

def replace(l,index,values):
    a=0
    for i in index:
        l[i]=values[a]
        a=+1
    return l

##############################################
# Forces and lengths:
##############################################

def ith_vertex_coords(ni, vertices):
    return vertices[ni]

def edge3D_length(el_nd_coords):
    [[x1, y1, z1], [x2, y2, z2]] = el_nd_coords
    return sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1) + (z2 - z1) * (z2 - z1))

def edge_vertex_indices(el):
    return el

def edge_vertex_coords(el, vertices):
    return [ith_vertex_coords(i, vertices) for i in edge_vertex_indices(el)]

def list_of_edge_lengths(edges, vertices):
    return [edge3D_length(edge_vertex_coords(el, vertices)) for el in edges]

def list_of_edge_forces(lengths, q):
    return list(map(lambda x, y: x * y, lengths, q))

def unstrained_length(l, f, ae):
    return ae * l / (ae + f)

def list_of_unstrained_lengths(ll, ff, ae):
    return list(map(lambda x, y, z: z * x / (z + y), ll, ff, ae))

##############################################
# Reactions:
##############################################

def nodal_resultant(nidx, tnei, vertices, edges, forces):
    eis = tnei[nidx]
    nc = vertices[nidx]
    sumf = zero_m(3)
    for ei in eis:
        ni2 = edges[ei][0]
        if ni2 == nidx:
            ni2 = edges[ei][1]
        nc2 = vertices[ni2]
        ee = nc2 - nc
        ee0 = ee / v_norm(ee)
        sumf = sumf + forces[ei] * ee0
    return sumf

#external load is not included in reactions to visualise the loads (unbalance force in loaded nodes) with draw_reactions function in GH
def list_of_nodal_resultants (vidx, tnei, vertices, edges, forces) :
    return [nodal_resultant(ni, tnei, vertices, edges, forces) for ni in vidx]

##############################################
##  Constraints:
##############################################

def map_value_to_all_edges(edges,v):
    return [v for x in range(len(edges))]

def edge_constraints(ed_indices, ed_values):
    constrs = []
    constrs.extend(zip(ed_indices, ed_values))
    return constrs

def add_edge_constraints(el_indices, el_values, constrs):
    constrs.extend(zip(el_indices, el_values))
    return constrs

def remove_edge_constraints(el_indices, constrs):
    constrs2 = []
    for cnst in constrs:
        ei = cnst[0]
        if ei not in el_indices:
            constrs2.append(cnst)
    return constrs2

##############################################
# Degress of freedom:
##############################################

def support_vertex_index(fixed):
    return fixed

def table_of_nodal_DsOF(v, fixed):
    dof_per_vertex = 1
    dof_table = zero_m((v, dof_per_vertex), dtype=int)
    for s in fixed:
        dof_table[support_vertex_index(s)] = -1
    ndof = 0
    j = 0
    for i in range(v):
        if dof_table[i, j] != -1:
            dof_table[i, j] = ndof
            ndof += 1
    return (ndof, dof_table)

def edge_DsOF(el_nd_ids, doft):
    nn = len(el_nd_ids)
    nd = len(doft[0])
    ix = zero_m((nn, nd), dtype=int)
    for i in range(nn):
        ix[i] = doft[el_nd_ids[i]]
    return ix.flatten()

##############################################
# Dead load:
##############################################

def enum_edges(edges):
    enum1 = 0
    enum_e = []
    for i in edges:
        enum_e.append([enum1, i])
        enum1 += 1
    return enum_e

def enum_vertices(vertices):
    enum1 = 0
    enum_v = []
    for i in vertices:
        enum_v.append([enum1, i])
        enum1 += 1
    return enum_v

def dead_load(uw, vertices, edges, fixed):
    l = list_of_edge_lengths(edges, vertices)
    weight = [i*j for i,j in zip(l, uw)]
    dead_load = [i * 0.5 for i in weight]
    dead_load_indexed = []
    enum_edg = enum_edges(edges)
    enum_nds = enum_vertices(vertices)
    for i in range(len(enum_edg)):
        dead_load_indexed.append([enum_edg[i][0], dead_load[i]])
    vertex_load_part = []
    for i in range(len(enum_nds)):
        for j in range(len(enum_edg)):
            if enum_nds[i][0] in enum_edg[j][1]:
                vertex_load_part.append([i, [0, 0, dead_load_indexed[j][1]]])
    vertex_load = []
    for i in range(len(enum_nds)):
        vertex_load.append([enum_nds[i][0], [0, 0, 0]])
    for i in range(len(vertex_load)):
        for j in range(len(vertex_load_part)):
            if vertex_load[i][0] == vertex_load_part[j][0]:
                vertex_load[i][1][2] += vertex_load_part[j][1][2]
    free_vertex_load = []
    for i in range(len(vertex_load)):
        if vertex_load[i][0] not in fixed:
            free_vertex_load.append(vertex_load[i])
    free_vertex_load_only = []
    for load in free_vertex_load:
        free_vertex_load_only.append(load[1])
    free_vertex_load_array = array_m(free_vertex_load_only)
    p_dead = as_matrix(free_vertex_load_array)
    return p_dead

def out_kicker(f,q):
    ind_q= [i for i, x in enumerate(f) if x <= 1e-03]
    for i in ind_q:
        q[i] = 0
    return q

##############################################
# Force Density Method:
##############################################

def mDmDf(ndof, doft, vertices, edges, q):
    ii = []
    jj = []
    aij = []
    mDf = zero_m((ndof, 3))
    for j in range(len(edges)):
        eni = edges[j]
        i0, i1 = edge_DsOF(eni, doft)
        if i0 != -1:
            ii.append(i0)
            jj.append(i0)
            aij.append(q[j])
            if i1 != -1:
                ii.append(i0)
                jj.append(i1)
                aij.append(-q[j])
            else:
                nc = vertices[eni[1]]
                mDf[i0, 0] -= q[j] * nc[0]
                mDf[i0, 1] -= q[j] * nc[1]
                mDf[i0, 2] -= q[j] * nc[2]
        if i1 != -1:
            ii.append(i1)
            jj.append(i1)
            aij.append(q[j])
            if i0 != -1:
                ii.append(i1)
                jj.append(i0)
                aij.append(-q[j])
            else:
                nc = vertices[eni[0]]
                mDf[i1, 0] -= q[j] * nc[0]
                mDf[i1, 1] -= q[j] * nc[1]
                mDf[i1, 2] -= q[j] * nc[2]
    mD = csc_matrix((aij, (ii, jj)), shape=(ndof, ndof))
    return mD, mDf

def table_of_vertex_edge_incidences(vertices, edges):
    nds_els = [[] for nd in vertices]
    for j in range(len (edges)):
        i0, i1 = edge_vertex_indices(edges[j])
        nds_els[i0].append(j)
        nds_els[i1].append(j)
    return nds_els

def __solve_lin_syst_d(mD, mDf):
    mDlu = splu(mD)
    cord = mDlu.solve(mDf)
    return cord

def table_of_nodal_coordinates(cord, vertices, doft, ndc=None):
    if ndc is None:
        ndc = zero_m((doft.shape[0], cord.shape[1]))
        for i in range(doft.shape[0]):
            if doft[i, 0] == -1:
                ndc[i] = vertices[i]
    for i in range(doft.shape[0]):
        if doft[i, 0] != -1:
            ndc[i] = cord[doft[i, 0]]
    return ndc

def nodal_coordinates_d(mD, mDf, vertices, doft):
    cc = __solve_lin_syst_d(mD, -mDf)
    ndc = table_of_nodal_coordinates(cc, vertices, doft)
    return ndc

def __solve_lin_syst_it(mD, mDf, x0, lin_solver, i_tol=1e-5, cord=None, callbacks=no_callbacks):
    (xx, infox) = lin_solver(mD, mDf[:, 0], x0[:, 0], tol=i_tol, callback=callbacks[0])
    (yy, infoy) = lin_solver(mD, mDf[:, 1], x0[:, 1], tol=i_tol, callback=callbacks[1])
    (zz, infoz) = lin_solver(mD, mDf[:, 2], x0[:, 2], tol=i_tol, callback=callbacks[2])
    if cord is None:
        cord = empty_m(mDf.shape)
    cord[:, 0] = xx
    cord[:, 1] = yy
    cord[:, 2] = zz
    return cord

def nodal_coordinates_it(mD, mDf, x0, vertices, doft, lin_solver=cg, i_tol=1e-5, callbacks=no_callbacks):
    cc = __solve_lin_syst_it(mD, -mDf, x0, lin_solver, i_tol, None, callbacks)
    ndc = table_of_nodal_coordinates(cc, vertices, doft)
    return ndc

def _FDM_load_d(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp):
     mD, mDf = mDmDf(ndof, tdof, vertices, edges, q)
     xyz = nodal_coordinates_d(mD, -p_dead_inp -p[free] + mDf, vertices, tdof)
     l = list_of_edge_lengths(edges, xyz)
     f = list_of_edge_forces(l, q)
     vidx = [vertices.index(i) for i in vertices]
     r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
     r = [list(x) for x in r]
     return xyz, f, l, r

def _FDM_load_it(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp, lin_solver, i_tol, callbacks):
    mD, mDf = mDmDf(ndof, tdof, vertices, edges, q)
    x0 = zero_m(mDf.shape)
    xyz = nodal_coordinates_it(mD, -p_dead_inp -p[free] + mDf, x0, vertices, tdof, lin_solver, i_tol, callbacks)
    l = list_of_edge_lengths(edges, xyz)
    f = list_of_edge_forces(l, q)
    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
    r = [list(x) for x in r]
    return xyz, f, l, r

def FDM(vertices, edges, fixed, q, loads, uw, lin_solver=splu, i_tol=1e-5, callbacks=no_callbacks):
    v = len(vertices)
    free = list(set(range(v)) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    ndof, tdof = table_of_nodal_DsOF(v, fixed)
    tnei = table_of_vertex_edge_incidences(vertices, edges)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    if lin_solver == splu:
        return _FDM_load_d(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp)
    else:
        return _FDM_load_it(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp, lin_solver, i_tol, callbacks)


def _FDM_load_dx(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp):
    mD, mDf = mDmDf(ndof, tdof, vertices, edges, q)
    xyz = nodal_coordinates_d(mD, -p_dead_inp - p[free] + mDf, vertices, tdof)
    return xyz


##############################################
# FDM multistep:
##############################################

def multistepFDM_d(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], steps=250, lin_solver=splu,
                callbacks=no_callbacks, cond_num=False):
    v = len(vertices)
    free = list(set(range(v)) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    tnei = table_of_vertex_edge_incidences(vertices, edges)
    q0 = copy(q)
    ndof, tdof = table_of_nodal_DsOF(v, fixed)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    xyz = _FDM_load_dx(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp)

    l = list_of_edge_lengths(edges, xyz)
    f = list_of_edge_forces(l, q0)

    for i in range(2, steps + 1):
        for fj in fcs:
            q0[fj[0]] = fj[1] / l[fj[0]]
        for lj in lcs:
            q0[lj[0]] = f[lj[0]] / lj[1]
        for l0j in l0cs:
            i0 = l0j[0]
            l0 = l0j[1][0]
            ae0 = l0j[1][1]
            ffj = f[i0]
            ll = (ae0 + ffj) * l0 / ae0
            q0[i0] = ffj / ll
        p_dead = dead_load(uw, xyz, edges, fixed)
        xyz = _FDM_load_dx(ndof, tdof, tnei, vertices, edges, q0, p, free, p_dead)
        l = list_of_edge_lengths(edges, xyz)
        f = list_of_edge_forces(l, q0)


    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
    r = [list(x) for x in r]

    return (xyz, f, q0, l, r)

def multistepFDM_it(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], steps=250, lin_solver=cg, i_tol=1e-5,
                callbacks=no_callbacks, cond_num=False):
    v = len(vertices)
    free = list(set(range(v)) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    tnei = table_of_vertex_edge_incidences(vertices, edges)
    q0 = copy(q)
    ndof, tdof = table_of_nodal_DsOF(v, fixed)
    mD, mDf = mDmDf(ndof, tdof, vertices, edges, q0)
    # if cond_num == True:
    #     print ('01', '  max =', mD.max(), '  nrm1 =', sm_norm(inv(mD), 1), '  nrmF =', sm_norm(inv(mD)), '  cn =', cond(
    #         mD.A, 1))
    x0 = zero_m(mDf.shape)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    cc = __solve_lin_syst_it(mD, p_dead_inp + p[free] - mDf, x0, lin_solver, i_tol, None, callbacks)
    xyz = table_of_nodal_coordinates(cc, vertices, tdof)

    l = list_of_edge_lengths(edges, xyz)
    f = list_of_edge_forces(l, q0)

    for i in range(2, steps + 1):
        for fj in fcs:
            q0[fj[0]] = fj[1] / l[fj[0]]
        for lj in lcs:
            q0[lj[0]] = f[lj[0]] / lj[1]
        for l0j in l0cs:
            i0 = l0j[0]
            l0 = l0j[1][0]
            ae0 = l0j[1][1]
            ffj = f[i0]
            ll = (ae0 + ffj) * l0 / ae0
            q0[i0] = ffj / ll
        mD, mDf = mDmDf(ndof, tdof, vertices, edges, q0)
        # if i % 10 == 1:
        #     print (i, '  max =', mD.max(), '  nrm1 =', sm_norm(inv(mD), 1), '  nrmF =', sm_norm(inv(mD)), '  cn =', cond(
        #         mD.A, 1))
        p_dead = dead_load(uw, xyz, edges, fixed)
        cc = __solve_lin_syst_it(mD, p_dead + p[free] - mDf, x0, lin_solver, i_tol, None, callbacks)
        xyz = table_of_nodal_coordinates(cc, vertices, tdof)
        l = list_of_edge_lengths(edges, xyz)
        f = list_of_edge_forces(l, q0)
    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
    r = [list(x) for x in r]

    return (xyz, f, q0, l, r)


# def multistepFDM(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], steps=250, lin_solver=splu, i_tol=1e-5,
#                 callbacks=no_callbacks, cond_num=False):
#     v = len(vertices)
#     free = list(set(range(v)) - set(fixed))
#     p = asarray(loads, dtype=float).reshape((-1, 3))
#     tnei = table_of_vertex_edge_incidences(vertices, edges)
#     if lin_solver == splu:
#         return _multistepFDM_load_d(tnei, vertices, edges, fixed, q, v, p, uw, free, fcs, lcs, l0cs, steps)
#     else:
#         return _multistepFDM_load_it(tnei, vertices, edges, fixed, q,v, p, uw, free, fcs, lcs, l0cs, steps,
#                                     lin_solver, i_tol, callbacks, cond_num)

##############################################
# FDM multistep with tolerance:
##############################################

def multistepFDM_wtol_d(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], tol_f=1e-3, tol_l=1e-3, tol_l0=1e-3, abs_error=True, lin_solver=splu,
                        latest=True, steps=250, callbacks=no_callbacks):
    v=len(vertices)
    free = list(set(range(v)) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    tnei = table_of_vertex_edge_incidences(vertices, edges)

    if abs_error == False:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]) / ff[1])
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]) / ll[1])
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0) / l0)
            return max(dl)

    else:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]))
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]))
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0))
            return max(dl)

    q0 = copy(q)
    ndof, tdof = table_of_nodal_DsOF(v, fixed)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    xyz = _FDM_load_dx(ndof, tdof, tnei, vertices, edges, q, p, free, p_dead_inp)

    l = list_of_edge_lengths(edges, xyz)
    f = list_of_edge_forces(l, q0)

    rfd = rld = rl0d = 0
    if fcs != []:
        rfd = force_diff(f, fcs)
    if lcs != []:
        rld = length_diff(l, lcs)
    if l0cs != []:
        rl0d = l0_diff(l, f, l0cs)
    # print (1, rfd, rld, rl0d)

    if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
        # print ('steps:', 1)
        # if fcs != []:
        #     print ('maximal force error: ', rfd)
        # if lcs != []:
        #     print ('maximal length error: ', rld)
        # if l0cs != []:
        #     print ('maximal unstrained length error: ', rl0d)
        return (xyz, f, q0)

    for i in range(2, steps + 1):
        if rfd >= tol_f:
            for fj in fcs:
                q0[fj[0]] = fj[1] / l[fj[0]]
        if rld >= tol_l:
            for lj in lcs:
                q0[lj[0]] = f[lj[0]] / lj[1]
        if rl0d >= tol_l0:
            for l0j in l0cs:
                i0 = l0j[0]
                l0 = l0j[1][0]
                ae0 = l0j[1][1]
                ffj = f[i0]
                ll = (ae0 + ffj) * l0 / ae0
                q0[i0] = ffj / ll

        p_dead = dead_load(uw, xyz, edges, fixed)
        xyz = _FDM_load_dx(ndof, tdof, tnei, vertices, edges, q0, p, free, p_dead)
        l = list_of_edge_lengths(edges, xyz)
        f = list_of_edge_forces(l, q0)

        if fcs != []:
            rfd = force_diff(f, fcs)
        if lcs != []:
            rld = length_diff(l, lcs)
        if l0cs != []:
            rl0d = l0_diff(l, f, l0cs)
        if i % 10 == 1:
            print (i, rfd, rld)

        if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
            break

    # if i == steps:
    #     print ('WARNING: maximal number of steps (' + str(steps) + ') reached !')
    # else:
    #     print ('steps:', i)
    # if fcs != []:
    #     print ('maximal force error: ', rfd)
    # if lcs != []:
    #     print ('maximal length error: ', rld)
    # if l0cs != []:
    #     print ('maximal unstrained length error: ', rl0d)

    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
    r = [list(x) for x in r]

    return (xyz, f, q0, l, r, i, rfd, rld, rl0d)

def multistepFDM_wtol_it(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], tol_f=1e-3, tol_l=1e-3, tol_l0=1e-3, abs_error=True, lin_solver=cg,
                      i_tol=None, latest=True, steps=250, callbacks=no_callbacks):
    v=len(vertices)
    free = list(set(range(v)) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    tnei = table_of_vertex_edge_incidences(vertices, edges)

    if abs_error == False:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]) / ff[1])
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]) / ll[1])
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0) / l0)
            return max(dl)

    else:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]))
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]))
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0))
            return max(dl)

    q0 = copy(q)
    ndof, tdof = table_of_nodal_DsOF(v, fixed)
    mD, mDf = mDmDf(ndof, tdof, vertices, edges, q0)
    x0 = zero_m(mDf.shape)
    nb = 5e-3
    i_tol = i_tol if i_tol != None else min(tol_f, tol_l, tol_l0) * nb
    # print ('i_tol =', i_tol)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    cc = __solve_lin_syst_it(mD, p_dead_inp + p[free] - mDf, x0, lin_solver, i_tol, None, callbacks)
    xyz = table_of_nodal_coordinates(cc, vertices, tdof)

    l = list_of_edge_lengths(edges, xyz)
    f = list_of_edge_forces(l, q0)

    rfd = rld = rl0d = 0
    if fcs != []:
        rfd = force_diff(f, fcs)
    if lcs != []:
        rld = length_diff(l, lcs)
    if l0cs != []:
        rl0d = l0_diff(l, f, l0cs)
    # print (1, rfd, rld, rl0d)

    if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
        # print ('steps:', 1)
        # if fcs != []:
        #     print ('maximal force error: ', rfd)
        # if lcs != []:
        #     print ('maximal length error: ', rld)
        # if l0cs != []:
        #     print ('maximal unstrained length error: ', rl0d)
        return (xyz, f, q0)

    for i in range(2, steps + 1):
        if rfd >= tol_f:
            for fj in fcs:
                q0[fj[0]] = fj[1] / l[fj[0]]
        if rld >= tol_l:
            for lj in lcs:
                q0[lj[0]] = f[lj[0]] / lj[1]
        if rl0d >= tol_l0:
            for l0j in l0cs:
                i0 = l0j[0]
                l0 = l0j[1][0]
                ae0 = l0j[1][1]
                ffj = f[i0]
                ll = (ae0 + ffj) * l0 / ae0
                q0[i0] = ffj / ll

        mD, mDf = mDmDf(ndof, tdof, vertices, edges, q0)
        x1 = cc if latest else x0
        p_dead = dead_load(uw, xyz, edges, fixed)
        cc = __solve_lin_syst_it(mD, p_dead+p[free]-mDf, x1, lin_solver, i_tol, None, callbacks)
        xyz = table_of_nodal_coordinates(cc, vertices, tdof)
        l = list_of_edge_lengths(edges, xyz)
        f = list_of_edge_forces(l, q0)

        if fcs != []:
            rfd = force_diff(f, fcs)
        if lcs != []:
            rld = length_diff(l, lcs)
        if l0cs != []:
            rl0d = l0_diff(l, f, l0cs)
        # if i % 10 == 1:
        #     print (i, rfd, rld, rl0d)

        if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
            break

    # if i == steps:
    #     print ('WARNING: maximal number of steps (' + str(steps) + ') reached !')
    # else:
    #     print ('steps:', i)
    # if fcs != []:
    #     print ('maximal force error: ', rfd)
    # if lcs != []:
    #     print ('maximal length error: ', rld)
    # if l0cs != []:
    #     print ('maximal unstrained length error: ', rl0d)

    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, xyz, edges, f)
    r = [list(x) for x in r]

    return (xyz, f, q0, l, r, i, rfd, rld, rl0d, i_tol)

# def multistepFDM_wtol(vertices, edges, fixed, q, loads, uw, fcs=[], lcs=[], l0cs=[], tol_f=1e-3, tol_l=1e-3, tol_l0=1e-3, abs_error=True, lin_solver=splu,
#                       i_tol=None, latest=True, steps=250, callbacks=no_callbacks):
#     v=len(vertices)
#     free = list(set(range(v)) - set(fixed))
#     p = asarray(loads, dtype=float).reshape((-1, 3))
#     tnei = table_of_vertex_edge_incidences(vertices, edges)
#     if lin_solver == splu:
#         return _multistepFDM_wtol_load_d(tnei, vertices, edges, fixed, q, v, p, uw, free, fcs, lcs, l0cs,
#                                     tol_f, tol_l, tol_l0, abs_error, steps)
#     else:
#         return _multistepFDM_wtol_load_it(tnei, vertices, edges, fixed, q, v, p, uw, free, fcs, lcs, l0cs, tol_f, tol_l, tol_l0,
#                                      abs_error, lin_solver, i_tol, latest, steps, callbacks)

##################################################
##Inexact Iterated FDM (tolerance optimisation):
##################################################

def multistepFDM_inexact_grid(vertices, edges, fixed, qs, loads, uw, fcs=[], lcs=[], l0cs=[],
                          tol_f=1e-3, tol_l=1e-3, tol_l0=1e-3, abs_error=True, lin_solver=cg,
                          i_tol_min=None, i_tol_max=0.1, damping=1e-2,
                          latest=True, steps=250, callbacks=no_callbacks):

    if abs_error == False:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]) / ff[1])
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]) / ll[1])
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0) / l0)
            return max(dl)

    else:

        def force_diff(f, fs):
            df = []
            for ff in fs:
                df.append(abs(f[ff[0]] - ff[1]))
            return max(df)

        def length_diff(l, ls):
            dl = []
            for ll in ls:
                dl.append(abs(l[ll[0]] - ll[1]))
            return max(dl)

        def l0_diff(l, f, l0s):
            dl = []
            for ll in l0s:
                i0 = ll[0]
                l0 = ll[1][0]
                ae0 = ll[1][1]
                dl.append(abs(unstrained_length(l[i0], f[i0], ae0) - l0))
            return max(dl)

    free = list(set(range(len(vertices))) - set(fixed))
    p = asarray(loads, dtype=float).reshape((-1, 3))
    qs0 = copy(qs)
    ndof, tdof = table_of_nodal_DsOF(len(vertices), fixed)
    mD, mDf = mDmDf(ndof, tdof, vertices, edges, qs0)
    x0 = zero_m(mDf.shape)
    nb = 5e-3
    i_tol0 = i_tol_min if i_tol_min != None else min(tol_f, tol_l, tol_l0) * nb
    # print ('i_tol_min =', i_tol0)
    p_dead_inp = dead_load(uw, vertices, edges, fixed)
    cc = __solve_lin_syst_it(mD, p_dead_inp + p[free] - mDf, x0, lin_solver, i_tol0, None, callbacks)
    nc = table_of_nodal_coordinates(cc, vertices, tdof)

    l = list_of_edge_lengths(edges, nc)
    f = list_of_edge_forces(l, qs0)

    rfd = rld = rl0d = 1e-13
    if fcs != []:
        rfd1 = rfd
        rfd = force_diff(f, fcs)
    if lcs != []:
        rld1 = rld
        rld = length_diff(l, lcs)
    if l0cs != []:
        rl0d1 = rl0d
        rl0d = l0_diff(l, f, l0cs)
    # print(1, i_tol0, rfd, rld, rl0d)

    if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
        # print ('steps:', 1)
        # if fcs != []:
        #     print ('maximal force error: ', rfd)
        # if lcs != []:
        #     print ('maximal length error: ', rld)
        # if l0cs != []:
        #     print ('maximal unstrained length error: ', rl0d)
        return (nc, f, qs0)

    i_tol1 = i_tol_max
    gamma_f = i_tol0 * (1 - sqrt(tol_f)) / tol_f ** 2
    gamma_l = i_tol0 * (1 - sqrt(tol_l)) / tol_l **2
    gamma_l0 = i_tol0 * (1 - sqrt(tol_l0)) / tol_l0 ** 2

    for i in range(2, steps + 1):
        if rfd >= tol_f:
            for fj in fcs:
                qs0[fj[0]] = fj[1] / l[fj[0]]
        if rld >= tol_l:
            for lj in lcs:
                qs0[lj[0]] = f[lj[0]] / lj[1]
        if rl0d >= tol_l:
            for l0j in l0cs:
                i0 = l0j[0]
                l0 = l0j[1][0]
                ae0 = l0j[1][1]
                ffj = f[i0]
                ll = (ae0 + ffj) * l0 / ae0
                qs0[i0] = ffj / ll

        #qs0 = out_kicker(f, qs0)
        mD, mDf = mDmDf(ndof, tdof, vertices, edges, qs0)
        x1 = cc if latest else x0
        minrfd = minrld = minrl0d = 0.0
        if fcs != []:
            minrfd = min(damping * rfd ** 3 / rfd1 ** 2, gamma_f * rfd ** 2)
        if lcs != []:
            minrld = min(damping * rld ** 3 / rld1 ** 2, gamma_l * rld ** 2)
        if l0cs != []:
            minrl0d = min(damping * rl0d ** 3 / rl0d1 ** 2, gamma_l0 * rl0d ** 2)
        i_tol1 = min(i_tol_max, i_tol1, max(minrfd, minrld, minrl0d, i_tol0))
        p_dead = dead_load(uw, nc, edges, fixed)
        cc = __solve_lin_syst_it(mD, p_dead + p[free]-mDf, x1, lin_solver, i_tol1, None, callbacks)
        nc = table_of_nodal_coordinates(cc, vertices, tdof)
        l = list_of_edge_lengths(edges, nc)
        f = list_of_edge_forces(l, qs0)

        if fcs != []:
            rfd1 = rfd
            rfd = force_diff(f, fcs)
        if lcs != []:
            rld1 = rld
            rld = length_diff(l, lcs)
        if l0cs != []:
            rl0d1 = rl0d
            rl0d = l0_diff(l, f, l0cs)

        # if i % 10 == 1:
            #print (i, i_tol1, rfd, rld, rl0d)


        if rfd < tol_f and rld < tol_l and rl0d < tol_l0:
            if abs(i_tol1-i_tol0) < 1e-13:
                break
            else:
                i_tol1 += i_tol0
                i_tol1 /= 2

    # if i == steps:
    #     print ('WARNING: maximal number of steps (' + str(steps) + ') reached !')
    # else:
    #     print ('steps:', i)
    # if fcs != []:
    #     print ('maximal force error: ', rfd)
    # if lcs != []:
    #     print ('maximal length error: ', rld)
    # if l0cs != []:
    #     print ('maximal unstrained length error: ', rl0d)

    tnei = table_of_vertex_edge_incidences(vertices, edges)
    vidx = [vertices.index(i) for i in vertices]
    r = list_of_nodal_resultants(vidx, tnei, nc, edges, f)
    r = [list(x) for x in r]

    return(nc, f, qs0, l, r, i, rfd, rld, rl0d, i_tol0, i_tol1)

# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    print('Hello FALCON user')
