
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

try:
    import ctypes
    from ctypes import c_int
    from ctypes import c_float
    from ctypes import c_double
    from ctypes import POINTER
    from ctypes import CFUNCTYPE

except ImportError:
    import sys
    if 'ironpython' not in sys.version.lower():
        raise

HERE = os.path.dirname(__file__)

SO = os.path.join(HERE, 'ifdm_d.so')
DLL = os.path.join(HERE, 'ifdm_d.dll')


__author__    = ['Tom Van Mele', ]
__copyright__ = 'Copyright 2017 - Block Research Group, ETH Zurich'
__license__   = 'MIT License'
__email__     = 'vanmelet@ethz.ch'


__all__ = [
    'ifdm_d_cpp',  
]


# ==============================================================================
# oo interface
# ==============================================================================


class Array1D(object):
    """"""

    def __init__(self, data, dtype):
        self._data = data
        self._dtype = dtype
        self._cdata = self.__array()

    @property
    def data(self):
        return self._data

    @property
    def cdata(self):
        return self._cdata

    @property
    def pydata(self):
        return list(self.cdata)

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        m = len(self.data)
        return m

    @property
    def ctype(self):
        m = self.shape
        if self.dtype == 'int':
            return c_int * m
        if self.dtype == 'float':
            return c_float * m
        if self.dtype == 'double':
            return c_double * m
        raise NotImplementedError

    def __array(self):
        data = self.data
        array = self.ctype
        return array(*data)


class Array2D(object):
    """"""

    def __init__(self, data, dtype):
        self._data = data
        self._dtype = dtype
        self._cdata = self.__array()

    @property
    def data(self):
        return self._data

    @property
    def cdata(self):
        return self._cdata

    @property
    def pydata(self):
        m, n = self.shape
        data = [[self.cdata[i][j] for j in range(n)] for i in range(m)]
        return data

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        data = self.data
        m, n = len(data), len(data[0])
        assert all(len(row) == n for row in data), 'The specified shape does not match the provided data.'
        return m, n

    @property
    def ctype(self):
        m, n = self.shape
        if self.dtype == 'int':
            return POINTER(c_int) * m
        if self.dtype == 'float':
            return POINTER(c_float) * m
        if self.dtype == 'double':
            return POINTER(c_double) * m
        raise NotImplementedError

    def __array(self):
        m, n = self.shape
        array = self.ctype(*[Array1D(self.data[i], self.dtype).cdata for i in range(m)])
        return array


class Array3D(object):
    """"""

    def __init__(self, data, dtype):
        self._data = data
        self._dtype = dtype
        self._cdata = self.__array()

    @property
    def data(self):
        return self._data

    @property
    def cdata(self):
        return self._cdata

    @property
    def pydata(self):
        m, n, o = self.shape
        data = [[[self.cdata[i][j][k] for k in range(o)] for j in range(n)] for i in range(m)]
        return data

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        data = self.data
        m, n, o = len(data), len(data[0]), len(data[0][0])
        assert all(len(row) == n for row in data), 'The specified shape does not match the provided data.'
        for row in data:
            assert all(len(col) == o for col in row), 'The specified shape does not match the provided data.'
        return m, n, o

    @property
    def ctype(self):
        m, n, o = self.shape
        if self.dtype == 'int':
            return POINTER(POINTER(c_int)) * m
        if self.dtype == 'float':
            return POINTER(POINTER(c_float)) * m
        if self.dtype == 'double':
            return POINTER(POINTER(c_double)) * m
        raise NotImplementedError

    def __array(self):
        m, n, o = self.shape
        if self.dtype == 'int':
            row = POINTER(c_int) * n
            col = c_int * o
        elif self.dtype == 'float':
            row = POINTER(c_float) * n
            col = c_float * o
        elif self.dtype == 'double':
            row = POINTER(c_double) * n
            col = c_double * o
        else:
            raise NotImplementedError
        a = self.ctype()
        for i in range(m):
            a[i] = row()
            for j in range(n):
                a[i][j] = col()
                for k in range(o):
                    a[i][j][k] = self.data[i][j][k]
        return a


# ==============================================================================
# functional interface
# ==============================================================================


def shape(data):
    m, n = len(data), len(data[0])
    assert len(data) == m, 'The specified shape does not match the provided data.'
    assert all(len(row) == n for row in data), 'The specified shape does not match the provided data.'
    return m, n


def zeros():
    pass


def ones():
    pass


def eye():
    pass

# ==============================================================================
# fdm_cpp 
# ==============================================================================

def ifdm_d_cpp (
        vertices, edges, fixed, loads, uws, qs,
        fcs, lcs, l0cs, tol_f, tol_l, tol_l0, steps
) :
    
    try:
        lib = ctypes.cdll.LoadLibrary(SO)
        print (lib)
    except Exception:
        try:
            lib = ctypes.cdll.LoadLibrary(DLL)
            print (DLL)
        except Exception:
            print(SO)
            print(DLL)
            raise

    fs = [0. for i in range (len (edges))]
    ls = [0. for i in range (len (edges))]
    rs = [[0., 0., 0.] for i in range (len (vertices))]
    d = [0.]

    if len(fcs) == 0:
        ifcs = []
        vfcs = []
    else:
        ifcs = [fcs[i][0] for i in range (len (fcs))]
        vfcs = [fcs[i][1] for i in range (len (fcs))]
    
    if len(lcs) == 0:
        ilcs = []
        vlcs = []
    else:   
        ilcs = [lcs[i][0] for i in range (len (lcs))]
        vlcs = [lcs[i][1] for i in range (len (lcs))]
        
    if len(l0cs) == 0:
        il0cs = []
        vl0cs = [[]]
    else:
        il0cs = [l0cs[i][0] for i in range (len (l0cs))]
        vl0cs = [l0cs[i][1] for i in range (len (l0cs))]

    cvertices  = Array2D (vertices, 'double')
    cedges     = Array2D (edges, 'int')
    cfixed     = Array1D (fixed, 'int')
    cloads     = Array2D (loads, 'double')
    cuws       = Array1D (uws, 'double')
    cqs        = Array1D (qs, 'double')

    cifcs      = Array1D (ifcs, 'int')
    cfcs       = Array1D (vfcs, 'double')
    cilcs      = Array1D (ilcs, 'int')
    clcs       = Array1D (vlcs, 'double')
    cil0cs     = Array1D (il0cs, 'int')
    cl0cs      = Array2D (vl0cs, 'double')
    
    clengths   = Array1D (ls, 'double')
    cforces    = Array1D (fs, 'double')
    cresiduals = Array2D (rs, 'double')
    cduration  = Array1D (d, 'double')

    tol_z = 1e-3

    lib.ifdm_d.argtypes = [
        ctypes.c_int, cvertices.ctype,
        ctypes.c_int, cfixed.ctype,
        ctypes.c_int, cedges.ctype, cqs.ctype,
        ctypes.c_int, cloads.ctype,
        ctypes.c_int, cuws.ctype,
        ctypes.c_int, cifcs.ctype, cfcs.ctype, 
        ctypes.c_int, cilcs.ctype, clcs.ctype, 
        ctypes.c_int, cil0cs.ctype, cl0cs.ctype,
        ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.c_double, ctypes.c_int, 
        cforces.ctype, clengths.ctype,
        cresiduals.ctype, cduration.ctype
    ]

    lib.ifdm_d (
        ctypes.c_int (len (vertices)), cvertices.cdata,
        ctypes.c_int (len (fixed)), cfixed.cdata,
        ctypes.c_int (len (edges)), cedges.cdata, cqs.cdata,
        ctypes.c_int (len (loads)), cloads.cdata,
        ctypes.c_int (len (uws)), cuws.cdata,
        ctypes.c_int (len (ifcs)), cifcs.cdata, cfcs.cdata,
        ctypes.c_int (len (ilcs)), cilcs.cdata, clcs.cdata,
        ctypes.c_int (len (il0cs)), cil0cs.cdata, cl0cs.cdata,
        tol_f, tol_l, tol_l0, tol_z, steps, 
        cforces.cdata, clengths.cdata,
        cresiduals.cdata, cduration.cdata
    )

    return cvertices.pydata, cforces.pydata, clengths.pydata, cresiduals.pydata, cduration.pydata


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass

